!-----------------------------------------------------------------------
!**BEGIN PROLOGUE Construct_Toeplitz
!**DATE WRITTEN   1998/10/01  (year/month/day)
!**AUTHORS
!
!  Marc Van Barel, Peter Kravanja
!
!    Department of Computer Science
!    Katholieke Universiteit Leuven
!    Celestijnenlaan 200 A
!    B-3001 Heverlee
!    BELGIUM
!    E-mail: Marc.VanBarel@cs.kuleuven.ac.be
!            Peter.Kravanja@na-net.ornl.gov
!
!**PURPOSE
!
! This module contains the subroutines CToeplitz and Toeplitz.
! CToeplitz enables the user to define a Toeplitz matrix by specifying
! the entries of a one-dimensional array.
! Toeplitz constructs the corresponding Toeplitz matrix.
!
!**END PROLOGUE Construct_Toeplitz
!-----------------------------------------------------------------------

MODULE Construct_Toeplitz

     USE Precision

     IMPLICIT NONE

!-----------------------------------------------------------------------
!**ACCESSABILITY
!
     PRIVATE
     PUBLIC :: CToeplitz, Toeplitz, nmatrix

!-----------------------------------------------------------------------
!**GLOBAL VARIABLE
!
     INTEGER :: nmatrix


CONTAINS


  SUBROUTINE CToeplitz(n,a,choice)
!-----------------------------------------------------------------------
!**PURPOSE
!
! Given the value of "choice" and "n", return a complex array of length
! 2n-1 that defines an n x n Toeplitz matrix.
!
!-----------------------------------------------------------------------
!  Parameters
!
! a DIMENSION(1:2*n-1)

  INTEGER, INTENT(IN)                         :: n, choice
  COMPLEX(KIND=DP), DIMENSION(:), INTENT(OUT) :: a

!-----------------------------------------------------------------------
!  Local variables
!
  REAL(KIND=DP), DIMENSION(:), ALLOCATABLE :: random
  INTEGER                                  :: k, nseed
  INTEGER, DIMENSION(:), ALLOCATABLE       :: seed

!-----------------------------------------------------------------------

  SELECT CASE (choice)

  CASE(1)
       CALL RANDOM_SEED(SIZE=nseed)
       ALLOCATE(seed(nseed))
       CALL RANDOM_SEED(GET=seed)
       seed(1) = seed(1) + nmatrix
       CALL RANDOM_SEED(PUT=seed)
       ALLOCATE(random(2*n-1))
       CALL RANDOM_NUMBER(random)
       a = random
       DEALLOCATE(random)
  CASE(2)
       a = 0.0_DP
       a(n) = 1.0_DP
       a(n-1) = 4.0_DP
       a(n+1) = 4.0_DP
  CASE(3)
       a = 0.0_DP
       DO k=1,n-1
         a(n-k)=(1.0_DP/k)**2
         a(n+k)=a(n-k)
       END DO
       a(n)=3.0_DP
  CASE(4)
       a = 0.0_DP
       a(n) = 1.0_DP
       a(n-1) = 4.0_DP
       a(n-2) = 5.0_DP
       a(n-3) = -1.0_DP
       a(n+1) = -27.0_DP
       a(n+2) = 1.0E-05_DP
  CASE(5)
       a = 0.0_DP
       DO k=1,n-1
         a(n-k)=(1.0_DP/k)**4
         a(n+k)=a(n-k)
       END DO
       a(n)=1.0_DP
  CASE(6)
       a = 0.0_DP
       DO k=1,n-1
         a(n-k)=(1.0_DP/k)**(1.3_DP)
         a(n+k)=a(n-k)
       END DO
       a(n)=1.0_DP
  CASE(7)
       a = 0.0_DP
       DO k=1,n-1
         a(n-k)=1.0_DP/k
         a(n+k)=a(n-k)
       END DO
       a(n)=0.0_DP

  END SELECT

  END SUBROUTINE CToeplitz
!
!-----------------------------------------------------------------------
!
     FUNCTION Toeplitz(n,t)
!-----------------------------------------------------------------------
!**PURPOSE
!
! Given an integer "n" and a complex array "t" of length 2n-1,
! return the corresponding n x n Toeplitz matrix.
!
!-----------------------------------------------------------------------
!  Parameters
!
       INTEGER, INTENT(IN)                        :: n
       COMPLEX(KIND=DP), INTENT(IN), DIMENSION(-n+1:n-1) :: t
       COMPLEX(KIND=DP), DIMENSION(n,n)           :: Toeplitz

!-----------------------------------------------------------------------
!  Local variables
!
       INTEGER                                    :: k, l

!-----------------------------------------------------------------------

       DO k = 1,n
         DO l = 1,n
           Toeplitz(k,l) = t(k-l)
         END DO
       END DO

     END FUNCTION Toeplitz
!
!-----------------------------------------------------------------------
!

END MODULE Construct_Toeplitz
