!-----------------------------------------------------------------------
!**BEGIN PROLOGUE RatInt_Module
!**DATE WRITTEN   1998/10/01  (year/month/day)
!**AUTHORS
!
!  Marc Van Barel, Peter Kravanja
!
!    Department of Computer Science
!    Katholieke Universiteit Leuven
!    Celestijnenlaan 200 A
!    B-3001 Heverlee
!    BELGIUM
!    E-mail: Marc.VanBarel@cs.kuleuven.ac.be
!            Peter.Kravanja@na-net.ornl.gov
!
!**PURPOSE
!
! This module contains subroutines for solving linearized rational interpolation
! problems. The main routines are RatIntSf (generically superfast algorithm
! for interpolation at roots of unity) and RatInt (fast-only) algorithm. 
!
!**END PROLOGUE RatInt_Module
!-----------------------------------------------------------------------

MODULE RatInt_Module

     USE Precision
     USE Timing
     USE Verbose_Module

     IMPLICIT NONE

!-----------------------------------------------------------------------
!**ACCESSABILITY
!
     PRIVATE
     PUBLIC   :: RatIntSf, RatInt, Gn, Gz, resL, resR
     PUBLIC   :: Glimit, Greflimit, Ginitscaling, Giscaling, Gfinalscaling
     PUBLIC   :: Gfactor, Gthreshold, Gnumzero, Gshowresiduals
     PUBLIC   :: Gfinalfactor, Gnitref, Gfactoritref, Gnfitref, Gffactoritref
     PUBLIC   :: Gdowndating

!-----------------------------------------------------------------------
!**GLOBAL VARIABLES

     INTEGER :: Gn
     COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE :: Gz
     COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE :: resL, resR
! Glimit indicates the level for which the RIP-problem is solved in a fast
! instead of a superfast way
     INTEGER :: Glimit
! Greflimit is an array indicating the levels at which downdating and iterative
! refinement is applied
     INTEGER, DIMENSION(:), ALLOCATABLE :: Greflimit
! Gshowresiduals = 0: no output of residuals for each Ratintfull
! Gshowresiduals = 1: compact output of residuals for each Ratintfull
! Gshowresiduals = 2: extended output of residuals for each Ratintfull
     INTEGER :: Gshowresiduals
! Giscaling = 0: no scaling used in recursive Ratintfull
! Giscaling = 1: global scaling used in recursive Ratintfull
! Giscaling = 2: global left-right scaling used in recursive Ratintfull
! Giscaling = 3: row scaling used in recursive Ratintfull
     INTEGER :: Ginitscaling, Giscaling, Gfinalscaling
! Gfactor indicates the factor used in recursive Ratintfull
     REAL(KIND=DP) :: Gfactor
! Gthreshold indicates the threshold used in Downdating
     REAL(KIND=DP) :: Gthreshold
! Gnumzero indicates the numerical zero used in Downdating
     REAL(KIND=DP) :: Gnumzero
! Gfinalfactor indicates the factor used in Ratintfull to handle the difficult
!   points
     REAL(KIND=DP) :: Gfinalfactor
! Gnitref indicates the maximum number of iterative refinement steps
     INTEGER :: Gnitref
! Gfactoritref indicates the improvement factor required in each step
! of iterative refinement
     REAL(KIND=DP) :: Gfactoritref
! Gnfitref indicates the maximum number of final iterative refinement steps
     INTEGER :: Gnfitref
! Gffactoritref indicates the improvement factor required in each step
! of final iterative refinement
     REAL(KIND=DP) :: Gffactoritref
! Gdowndating indicates if downdating has to be applied or not.
     LOGICAL :: Gdowndating

!-----------------------------------------------------------------------
      EXTERNAL dcffti, dcfftf, dcfftb


CONTAINS


      SUBROUTINE RatIntSf(resL,resR,m,tau,b,d,alpha,hn)
!-----------------------------------------------------------------------
!**PURPOSE
!
! Divide and conquer algorithm for generically superfast rational interpolation
! at roots of unity. Difficult points are handled at the end via the fast-only
! algorithm RatInt.
!
!-----------------------------------------------------------------------
!  Parameters
!
      INTEGER, INTENT(IN)                            :: m, tau
      INTEGER, INTENT(OUT)                           :: hn
      REAL(KIND=DP), INTENT(OUT)                     :: alpha
! INOUT because of the possible scaling of resL and resR
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT)  :: resL, resR
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(OUT)    :: b, d

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                              :: ndifficult, ndone, k
      INTEGER, DIMENSION(3)                :: state, stateout
      INTEGER, DIMENSION(m)                :: ipoints
      REAL(KIND=DP)                        :: time01, time02
      COMPLEX(KIND=DP), DIMENSION(m)       :: a, c
      COMPLEX(KIND=DP), DIMENSION(:), allocatable   :: Lrem, Rrem

!-----------------------------------------------------------------------


      IF (Ginitscaling == 1) THEN
        CALL GScaling(resL,resR)
      ELSE IF (Ginitscaling == 2) THEN
        CALL LRScaling(resL,resR,alpha)
      ELSE IF (Ginitscaling == 3) THEN
        CALL PScaling(resL,resR)
      END IF

      state = (/ tau, 1, 1-tau /)

      time01 = GiveTime()

        CALL RecRatInt(resL,resR,m,state,stateout,Glimit,  &
          Greflimit,0,1,a,b,c,d,ndifficult,ipoints)

      time02 = GiveTime()

      IF (verbose > 0) THEN
        print*,' '
        print*,'  Execution time of subroutine RecRatInt'
        print*,'  for ',m,' interpolation conditions = ',time02-time01
      END IF

!!      WRITE(94,'(A)') 'Execution time of subroutine RecRatInt ='
!!      WRITE(94,8010) time02-time01

!! kan de variable stateout weggelaten worden?
      state = stateout

      IF (verbose > 0) THEN
        print*,' '
        print *, '  number of difficult points = ', ndifficult
        print*,' '
      END IF
      WRITE(8,1011) ndifficult
1011  FORMAT(I10)

      
        IF (ndifficult > 0) THEN

          IF (verbose > 0) THEN
            print*,' '
            print *, '  *** BEGIN HANDLING DIFFICULT POINTS ***'
            print*,' '
          END IF


          time01 = GiveTime()

          ALLOCATE(Lrem(m)); ALLOCATE(Rrem(m))
!! too much work +/- (--> FFT <> bijv. Horner)
          CALL GetRes(a,b,c,d,resL,resR,m,.FALSE.,0,Lrem,Rrem)

          ndone = m - ndifficult

          CALL RatInt(Gz(ipoints(ndone+1:m)),Lrem(ipoints(ndone+1:m)),   &
            Rrem(ipoints(ndone+1:m)),ndifficult,Gfinalfactor,state,a,c,b,d,   &
            hn,ipoints(ndone+1:m),Gfinalscaling,0,.TRUE.)
          DEALLOCATE(Lrem,Rrem)

          hn = hn + ndone

          time02 = GiveTime()
          IF (verbose > 0) THEN
            PRINT *, '    Execution time to handle the difficult points = ',&
                          time02-time01
          END IF

          WRITE(94,'(A)') 'Execution time to handle the difficult points ='
          WRITE(94,8010) time02-time01

          IF (verbose > 0) THEN
            print*,' '
            print *, '  *** END HANDLING DIFFICULT POINTS ***'
            print*,' '
          END IF

        ELSE
          hn = m
        END IF

        IF (Gnfitref > 0) THEN
          ipoints = (/ (k, k=1,m) /)
          CALL ItRefSf(a,b,c,d,m,resL,resR,state,Gnfitref,Gffactoritref,&
                       ipoints,0)
        END IF

8010  FORMAT(1x,ES9.2)

      END SUBROUTINE RatIntSf 


!-----------------------------------------------------------------------


      SUBROUTINE RatInt(Iz,IL,IR,m,factor,state,        &
                        coeffNL,coeffNR,coeffDL,coeffDR,   &
                        hn,ipoints,iscaling,show,full)
!-----------------------------------------------------------------------
!**PURPOSE
!
!  Given the interpolation points Iz and the interpolation data
!  IL and IR, subroutine RatInt updates the basis matrix
!  represented by the four components coeffNL,coeffNR,coeffDL
!  and coeffDR. 
!
!  To enhance numerical stability pivoting is used, i.e.,
!  the interpolation conditions can be permuted.
!
!  ipoints contains the permuted indices of the interpolation conditions.
!
!  Also state = [tau degl  degr] is updated.
!
!  degl and degr are the tau-degrees (+1) of the left and right column
!  of the basis matrix.
!  
!  If "full" is .FALSE., only the denominator part of the basis
!  matrix coeffDL and coeffDR is updated.
!
!  iscaling = 0: no scaling
!  iscaling = 1: global scaling
!  iscaling = 2: global left-right scaling
!  iscaling = 3: row scaling
!
!  If all remaining residual elements which can be taken as pivot elements
!  are smaller than or equal to factor*EPSILON(factor), the subroutine
!  returns the basis matrix with the corresponding state-vector
!  as it is updated so far.
!  Note that when the IN-variable factor is zero, all interpolation
!  conditions are handled except if all remaining residuals are equal
!  to zero.
!  The output-variable hn indicates the number of interpolation conditions
!  handled.
!
!  On input the INOUT variable
!       coeffNL has to be at least degl+m elements,
!       coeffNR has to be at least degr+m elements,
!       coeffDL has to be at least degl+tau+m elements,
!       coeffDR has to be at least degr+tau+m elements.
!
!-----------------------------------------------------------------------
!  Parameters
!
! Iz, IL, IR DIMENSION(m)
      LOGICAL, INTENT(IN)                              :: full
      INTEGER, INTENT(IN)                              :: m, iscaling, show
      INTEGER, INTENT(OUT)                             :: hn 
      INTEGER, DIMENSION(:), INTENT(INOUT)             :: state
      INTEGER, DIMENSION(:), INTENT(OUT)               :: ipoints
      REAL(KIND=DP), INTENT(IN)                        :: factor 
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)       :: Iz, IL, IR
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT)    :: coeffNL, coeffNR
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT)    :: coeffDL, coeffDR

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                      :: degl, degr, degrN, deglN, degrD, deglD
      INTEGER                      :: tau
      INTEGER                      :: i, ipivL, ipivR, flow, ipiv
      REAL(KIND=DP)                :: reps, maxpivL, maxpivR
      REAL(KIND=DP)                :: alpha 
      COMPLEX(KIND=DP)             :: hold, mult, hh
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE :: z, L, R

!-----------------------------------------------------------------------

      ALLOCATE(z(m))
      ALLOCATE(L(m))
      ALLOCATE(R(m))
      z = Iz
      L = IL
      R = IR

!! testresiduals
      IF (show > 0) THEN
        CALL PlotRes(m,L,R,5,'residuals.dat',show)
      END IF
!! testresiduals

      IF (iscaling == 1) THEN
        CALL GScaling(L,R)
      ELSE IF (iscaling == 2) THEN
        CALL LRScaling(L,R,alpha)
      ELSE IF (iscaling == 3) THEN
        CALL PScaling(L,R)
      END IF

      tau  = state(1)
      degl = state(2)
      degr = state(3)

      reps = factor*EPSILON(factor)

      ipoints = (/ (i, i=1,m) /)

      DO i = 1, m 
       IF (degl == degr) THEN
           CALL V2Max(ABS(REAL(R(i:m))),ABS(AIMAG(R(i:m))),maxpivR,ipivR)
           ipivR = ipivR + i-1
           CALL V2Max(ABS(REAL(L(i:m))),ABS(AIMAG(L(i:m))),maxpivL,ipivL)
           ipivL = ipivL + i-1
           IF (maxpivR <= reps .AND. maxpivL <= reps) THEN
             hn = i-1
             state(1) = tau; state(2) = degl; state(3) = degr
             DEALLOCATE(z)
             DEALLOCATE(L)
             DEALLOCATE(R)
             RETURN
           END IF
           IF (maxpivR > maxpivL) THEN
             ipiv = ipivR; flow = 2
!! testpivot
             WRITE(9,9876) i,maxpivR
!! testpivot
           ELSE
             ipiv = ipivL; flow = 1
!! testpivot
             WRITE(9,9876) i,maxpivL
!! testpivot
           END IF
        ELSEIF (degl < degr) THEN
           CALL V2Max(ABS(REAL(L(i:m))),ABS(AIMAG(L(i:m))),maxpivL,ipivL)
           ipivL = ipivL + i-1
           IF (maxpivL <= reps) THEN
             hn = i-1
             state(1) = tau; state(2) = degl; state(3) = degr
             DEALLOCATE(z)
             DEALLOCATE(L)
             DEALLOCATE(R)
             RETURN
           END IF 
           ipiv = ipivL; flow = 1
!! testpivot
           WRITE(9,9876) i,maxpivL
!! testpivot
        ELSE
           CALL V2Max(ABS(REAL(R(i:m))),ABS(AIMAG(R(i:m))),maxpivR,ipivR)
           ipivR = ipivR + i-1
           IF (maxpivR <= reps) THEN
             hn = i-1
             state(1) = tau; state(2) = degl; state(3) = degr
             DEALLOCATE(z)
             DEALLOCATE(L)
             DEALLOCATE(R)
             RETURN
           END IF
           ipiv = ipivR; flow = 2
!! testpivot
           WRITE(9,9876) i,maxpivR
!! testpivot
        END IF

!! testpivot
9876    FORMAT(I8,1E16.3)
!! testpivot

       hold = z(i); z(i) = z(ipiv); z(ipiv) = hold 
       hold = L(i); L(i) = L(ipiv); L(ipiv) = hold 
       hold = R(i); R(i) = R(ipiv); R(ipiv) = hold 
       hold = ipoints(i); ipoints(i) = ipoints(ipiv); ipoints(ipiv) = hold

!-----------------------------------------------------------------------
!  flow = 1 : the left  column contains the pivot element
!  flow = 2 : the right column contains the pivot element
!-----------------------------------------------------------------------

       SELECT CASE(flow)
         CASE(1)
           mult = -R(i)/L(i)
           R(i+1:m) = R(i+1:m) + mult*L(i+1:m)

           deglD = degl + tau

           IF (deglD > 0) THEN
             coeffDR(1:deglD) = coeffDR(1:deglD) + mult*coeffDL(1:deglD)
             hh = coeffDL(deglD)
             coeffDL(2:deglD) = coeffDL(1:deglD-1) - z(i)*coeffDL(2:deglD)
             coeffDL(1) = - coeffDL(1)*z(i)
             coeffDL(deglD + 1) = hh
           END IF

           IF (FULL) THEN
             deglN = degl
             IF (deglN > 0) THEN
               coeffNR(1:deglN) = coeffNR(1:deglN) + mult*coeffNL(1:deglN)
               hh = coeffNL(deglN)
               coeffNL(2:deglN) = coeffNL(1:deglN-1) - z(i)*coeffNL(2:deglN)
               coeffNL(1) = - coeffNL(1)*z(i)
               coeffNL(deglN + 1) = hh
             END IF
           END IF

           degl = degl + 1

           L(i+1:m) = L(i+1:m) * (z(i+1:m)-z(i))

         CASE(2)
           mult = -L(i)/R(i) 
           L(i+1:m) = L(i+1:m) + mult*R(i+1:m)

           degrD = degr + tau

           IF (degrD > 0) THEN
             coeffDL(1:degrD) = coeffDL(1:degrD) + mult*coeffDR(1:degrD)
             hh = coeffDR(degrD)
             coeffDR(2:degrD) = coeffDR(1:degrD-1) - z(i)*coeffDR(2:degrD)
             coeffDR(1) = - coeffDR(1)*z(i)
             coeffDR(degrD + 1) = hh
           END IF

           IF (FULL) THEN
             degrN = degr
             IF (degrN > 0) THEN
               coeffNL(1:degrN) = coeffNL(1:degrN) + mult*coeffNR(1:degrN)
               hh = coeffNR(degrN)
               coeffNR(2:degrN) = coeffNR(1:degrN-1) - z(i)*coeffNR(2:degrN)
               coeffNR(1) = - coeffNR(1)*z(i)
               coeffNR(degrN + 1) = hh
             END IF
    
             degr = degr + 1
           END IF
  
           R(i+1:m) = R(i+1:m) * (z(i+1:m) - z(i))
       END SELECT

      END DO

      hn = m
      state(1) = tau; state(2) = degl; state(3) = degr

      IF (iscaling == 2) THEN
        coeffDL(1:degl+tau) = coeffDL(1:degl+tau) * alpha
        IF (FULL) THEN
          coeffNR(1+degl) = coeffNR(1+degl) * alpha
        ENDIF
      ENDIF

      DEALLOCATE(z)
      DEALLOCATE(L)
      DEALLOCATE(R)

      END SUBROUTINE RatInt


!-----------------------------------------------------------------------


      RECURSIVE SUBROUTINE RecRatInt(resL,resR,n,state,stateout, &
        limit,reflimit,Gi,Ginc,a,b,c,d,ndiff,ipoints)
!-----------------------------------------------------------------------
!  Parameters
!

      INTEGER, INTENT(IN)                        :: n, Gi, Ginc, limit
      INTEGER, DIMENSION(:), INTENT(IN)          :: reflimit
      INTEGER, INTENT(OUT)                       :: ndiff
      INTEGER, DIMENSION(:), INTENT(IN)          :: state
      INTEGER, DIMENSION(:), INTENT(OUT)         :: stateout
      INTEGER, DIMENSION(:), INTENT(OUT)         :: ipoints
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)   :: resL, resR
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(OUT)  :: a, b, c, d

!-----------------------------------------------------------------------
!  Local variables
!
      LOGICAL                         :: full, itref
      INTEGER(4) :: hGi
      INTEGER                         :: i, k, i1, i2
      INTEGER                         :: show
      INTEGER                         :: tau, hn, ndiv2
      INTEGER                         :: ndiff1, ndiff2 
      INTEGER                         :: degl2, degr2, degl, degr
      INTEGER                         :: tau1, degl1, degr1
      INTEGER                         :: deglN, degrN, deglD, degrD
      INTEGER                         :: ndone, n1done, n2done
      INTEGER                                      :: ndowndating
      INTEGER, DIMENSION(3)                        :: state1, state2
      INTEGER, DIMENSION(:), ALLOCATABLE           :: ipoints1, ipoints2
      REAL(KIND=DP), DIMENSION(:), ALLOCATABLE     :: wsave2
      REAL(KIND=DP)                                :: time01, time02
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: resL1, resR1
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: a1, b1, c1, d1
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: a2, b2, c2, d2 
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: a1eval2, b1eval2
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: c1eval2, d1eval2
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: pws, resL2, resR2
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: hhh

!-----------------------------------------------------------------------

      tau = state(1)

      hGi = Gi

      IF (tau <= -n) THEN
        a = 0.0_DP; a(1) = -Gz(MOD(hGi*n,Gn)+1)
        b = 0.0_DP; d = 0.0_DP; d(1) = 1.0_DP
        c = -resR / resL

        CALL PolInterp(c,n,Gi)

        ipoints = (/ (k,k=1,n) /)
        stateout(1) = tau
        stateout(2) = n+1
        stateout(3) = MAX(n,1-tau)
        ndiff = 0
        RETURN
      END IF
      
      IF (n <= 2**limit) THEN
        tau = state(1)
        a = 0.0_DP; c = 0.0_DP
        a(1) = 1.0_DP
        b = 0.0_DP; d = 0.0_DP
        d(1) = 1.0_DP
        stateout(1) = tau; stateout(2) = 1; stateout(3) = 1 - tau
        show = 0
        full = .TRUE.

        time01 = GiveTime()

        CALL RatInt(Gz(Gi+1:Gi+1+(n-1)*Ginc:Ginc),resL,resR, &
                    n,Gfactor,stateout,a,c,b,d,   &
                    hn,ipoints,Giscaling,show,full) 

        time02 = GiveTime()

        IF (verbose > 0) THEN
          print*,'    Execution time of subroutine RatInt = ',time02-time01
        END IF
!!        WRITE(94,'(A)') 'Execution time of subroutine RatInt = '
!!        WRITE(94,8010) time02-time01
        ndiff = n - hn
      ELSE
        ndiv2 = n/2

        ALLOCATE(resL1(ndiv2))
        ALLOCATE(resR1(ndiv2))
        resL1=resL(1:n:2)
        resR1=resR(1:n:2)
        ALLOCATE(a1(ndiv2))
        ALLOCATE(b1(ndiv2))
        ALLOCATE(c1(ndiv2))
        ALLOCATE(d1(ndiv2))
        ALLOCATE(ipoints1(ndiv2))
        ipoints1 = ipoints(1:ndiv2)
        CALL RecRatInt(resL1,resR1,ndiv2,state, &
                       state1,limit,  &
                       reflimit,Gi,2*Ginc,a1,b1,c1,  &
                       d1, &
                       ndiff1,ipoints1)

        DEALLOCATE(resL1)
        DEALLOCATE(resR1)


        ALLOCATE(a1eval2(ndiv2)); ALLOCATE(b1eval2(ndiv2))
        ALLOCATE(c1eval2(ndiv2)); ALLOCATE(d1eval2(ndiv2))

!! can be programmed differently??

        ALLOCATE(pws(ndiv2))

        CALL Powers(Gi+Ginc,ndiv2,pws) 

        tau1 = state1(1); degl1 = state1(2); degr1 = state1(3)
        ALLOCATE(wsave2(4*ndiv2+15))
        CALL dcffti(ndiv2,wsave2)

        IF (degl1 > ndiv2) THEN
          i1 = MOD((hGi+Ginc)*ndiv2,Gn)
          i2 = MOD(hGi*ndiv2,Gn)
          a1eval2 = Gz(i1+1)-Gz(i2+1)
          b1eval2 = 0.0_DP
        ELSE
          a1eval2 = a1; b1eval2 = b1
          a1eval2 = pws * a1eval2
          b1eval2 = pws * b1eval2

          a1eval2 = CONJG(a1eval2)
          b1eval2 = CONJG(b1eval2)
          CALL dcfftf(ndiv2,a1eval2,wsave2)
          CALL dcfftf(ndiv2,b1eval2,wsave2)
          a1eval2 = CONJG(a1eval2)
          b1eval2 = CONJG(b1eval2)
        END IF

        IF (degr1+tau1 > ndiv2) THEN
          i1 = MOD((hGi+Ginc)*ndiv2,Gn)
          i2 = MOD(hGi*ndiv2,Gn)
          d1eval2 = Gz(i1+1)-Gz(i2+1)
          c1eval2 = 0.0_DP
        ELSE
          c1eval2 = c1; d1eval2 = d1
          c1eval2 = pws * c1eval2
          d1eval2 = pws * d1eval2

          c1eval2 = CONJG(c1eval2)
          d1eval2 = CONJG(d1eval2)
          CALL dcfftf(ndiv2,c1eval2,wsave2)
          CALL dcfftf(ndiv2,d1eval2,wsave2)
          c1eval2 = CONJG(c1eval2)
          d1eval2 = CONJG(d1eval2)
        END IF

        DEALLOCATE(pws)
        DEALLOCATE(wsave2)

        ALLOCATE(resL2(ndiv2)); ALLOCATE(resR2(ndiv2))

        resL2 = resL(2:n:2) * a1eval2 + resR(2:n:2) * b1eval2
        resR2 = resL(2:n:2) * c1eval2 + resR(2:n:2) * d1eval2

        DEALLOCATE(a1eval2); DEALLOCATE(b1eval2)
        DEALLOCATE(c1eval2); DEALLOCATE(d1eval2)

        state1(1) = degl1 - degr1

        ALLOCATE(a2(ndiv2))
        ALLOCATE(b2(ndiv2))
        ALLOCATE(c2(ndiv2))
        ALLOCATE(d2(ndiv2))
        ALLOCATE(ipoints2(ndiv2))
        ipoints2 = ipoints(ndiv2+1:n)
        CALL RecRatInt(resL2,resR2,ndiv2,state1,state2,   &
          limit,reflimit,Gi+Ginc,2*Ginc,a2,b2, &
          c2,d2,ndiff2,ipoints2)

        DEALLOCATE(resL2); DEALLOCATE(resR2)

        ndiff = ndiff1 + ndiff2

        tau = tau1
        degl2 = state2(2); degr2 = state2(3)
        degl = degl1 + degl2 - 1
        degr = degl1 + degr2 - 1

        stateout(1) = tau; stateout(2) = degl; stateout(3) = degr

        n1done = ndiv2 - ndiff1; n2done = ndiv2 - ndiff2
        ndone = n1done + n2done

        ipoints1 = 2*ipoints1-1
        ipoints2 = 2*ipoints2
        ipoints(1:n1done) = ipoints1(1:n1done)
        ipoints(n1done+1:ndone) = ipoints2(1:n2done)
        ipoints(ndone+1:ndone+ndiff1) = ipoints1(n1done+1:ndiv2)
        ipoints(ndone+ndiff1+1:n) = ipoints2(n2done+1:ndiv2)
        DEALLOCATE(ipoints1); DEALLOCATE(ipoints2)

        CALL PolMatMult(a1,b1,c1,d1,a2,b2,c2,d2,state1,state2,n,a,b,c,d)

        DEALLOCATE(a1)
        DEALLOCATE(b1)
        DEALLOCATE(c1)
        DEALLOCATE(d1)
        DEALLOCATE(a2)
        DEALLOCATE(b2)
        DEALLOCATE(c2)
        DEALLOCATE(d2)

        deglN = MAX(degl,0); degrN = MAX(degr,0)
        deglD = MAX(degl + tau,0); degrD = MAX(degr + tau,0)
        IF (degl > n) THEN
          a = 0.0_DP
          a(n+1) = 1.0_DP
          a(1) = -Gz(MOD(hGi*n,Gn)+1)
          b = 0.0_DP
        ELSE
          a(deglN+1:n) = 0.0_DP
          b(deglD+1:n) = 0.0_DP
        END IF

        IF (degr+tau > n) THEN
          c = 0.0_DP
          d = 0.0_DP
          d(n+1) = 1.0_DP
          d(1) = -Gz(MOD(hGi*n,Gn)+1)
        ELSE
          c(degrN+1:n) = 0.0_DP
          d(degrD+1:n) = 0.0_DP
        END IF
      END IF

      itref = .FALSE.
      i = 1
      DO
        IF (itref.OR.(i > Size(reflimit))) EXIT
        itref = (n == 2**reflimit(i))
        i = i+1
      END DO

      IF (itref .AND. Gdowndating) THEN
        show = 2
        IF (verbose > 0) THEN
          print*,' '
          print*, '   *** BEGIN DOWNDATING ***'
          print*,' '
        END IF
        time01 = GiveTime()
        CALL Downdating(a,b,c,d,Gz(Gi+1:Gi+1+(n-1)*Ginc:Ginc), &
                        resL,resR,n,show,Gi,stateout, &
                        ipoints,ndowndating)
        time02 = GiveTime()
        IF (verbose > 0) THEN
          print*,'    Execution time for subroutine Downdating'
          print*,'    for ',n,' interpolation conditions = ',time02-time01
          print*,' '
          print*, '   *** END DOWNDATING ***'
          print*,' '
        END IF
        WRITE(94,'(A)') 'Execution time for subroutine Downdating ='
        WRITE(94,8010) time02-time01
        ndiff = ndiff + ndowndating
      END IF


      IF (itref) THEN
        tau = stateout(1); degl = stateout(2); degr = stateout(3)
        IF ( ( (degl < degr).AND. &
               (ABS(a(MAX(degl,1))) < ABS(b(MAX(degl+tau,1)))) ) &
             .OR. &
             ( (degr < degl).AND. &
               (ABS(c(MAX(degr,1))) > ABS(d(MAX(degr+tau,1)))) ) ) THEN
          ALLOCATE(hhh(n))
          hhh = a; a = c; c = hhh
          hhh = b; b = d; d = hhh
          DEALLOCATE(hhh)
          stateout(2) = degr; stateout(3) = degl
        END IF

!!        nnitref = NINT(log10(real(n))/log10(2.e0)) - reflimit + 1

        time01 = GiveTime()
        CALL ItRefSf(a,b,c,d,n,resL,resR,stateout,Gnitref,Gfactoritref,&
                     ipoints,Gi)
        time02 = GiveTime()
        IF (verbose > 0) THEN
          print*,'    Execution time for subroutine ItRefSf'
          print*,'    for ',n,' interpolation conditions = ',time02-time01
        END IF
        WRITE(94,'(A)') 'Execution time for subroutine ItRefSf ='
        WRITE(94,8010) time02-time01

      END IF

8010  FORMAT(2X,ES9.2)

      END SUBROUTINE RecRatInt


!-----------------------------------------------------------------------


      SUBROUTINE ItRefSf(a,b,c,d,npts,resL,resR,state,nitref,factor,ipoints,Gi)
!-----------------------------------------------------------------------
!  Parameters
!
! ipoints, resL, resR, a, b, c, d DIMENSION(npts)
      INTEGER, INTENT(IN)                             :: nitref, Gi, npts
      INTEGER, DIMENSION(3), INTENT(IN)               :: state
      INTEGER, DIMENSION(:), INTENT(IN)               :: ipoints
      REAL(KIND=DP), INTENT(IN)                       :: factor
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)      :: resL, resR
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT)   :: a, b, c, d

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER  :: tau, degl, degr, deglN, deglD, degrN, degrD, degN, degD
      INTEGER                                    :: n, m, capN, k, i, j, mn
      REAL(KIND=DP)                              :: wsave(4*npts+15)
      REAL(KIND=DP)                              :: wsave2(8*npts+15)
      REAL(KIND=DP)                              :: mrL, mrR,mresL,mresR
      REAL(KIND=DP)                              :: acc, pacc
      REAL(KIND=DP)                              :: time01, time02
      COMPLEX(KIND=DP)                             :: pt, mult, detB
      COMPLEX(KIND=DP)                             :: ma, mb, mc, md, detm
      COMPLEX(KIND=DP), DIMENSION(2,2)             :: Bmat
      COMPLEX(KIND=DP), DIMENSION(npts)            :: pws, h1, h2 
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: val11, val12
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: val21, val22
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: dval11, dval12
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: dval21, dval22
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: w11e, w12e
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: w21e, w22e
      COMPLEX(KIND=DP), DIMENSION(npts)            :: rL, rR
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: x1, x2, x1ae, x2ae
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE  :: ah, bh, ch, dh
      
!-----------------------------------------------------------------------

      time01 = GiveTime()

      IF (verbose > 0) THEN
        print*,' '
        print*,'   *** BEGIN ITERATIVE REFINEMENT ***'
        print*,' '
      END IF

      tau = state(1); degl = state(2); degr = state(3)

      deglN = degl; deglD = degl + tau; degrN = degr; degrD = degr + tau
      degN = MAX(deglN,degrN)
      degD = MAX(deglD,degrD)

      n = deglN - 1; m = degrD - 1

      IF (verbose > 0) THEN
        print*,'    The degree of the upper left element of '
        print*,'    the basis matrix is ', n,'.'
        print*,'    The degree of the lower right element of '
        print*,'    the basis matrix is ', m,'.'
        print*,' '
      END IF

      IF (degl < degr) THEN
        mult = a(deglN); a = a/mult; b = b/mult
        mult = -c(degrN)
        c(2:deglN+1) = c(2:deglN+1) + mult*a(1:deglN)
        d(2:deglD+1) = d(2:deglD+1) + mult*b(1:deglD)
        mult = -c(degrN-1)
        c(1:deglN) = c(1:deglN) + mult*a(1:deglN)
        d(1:deglD) = d(1:deglD) + mult*b(1:deglD)
        mult = d(degrD); d = d/mult; c = c/mult
      ELSE IF (degl > degr) THEN
        mult = d(degrD); d = d/mult; c = c/mult
        mult = -b(deglD)
        a(2:degrN+1) = a(2:degrN+1) + mult*c(1:degrN)
        b(2:degrD+1) = b(2:degrD+1) + mult*d(1:degrD)
        mult = -b(deglD-1)
        a(1:degrN) = a(1:degrN) + mult*c(1:degrN)
        b(1:degrD) = b(1:degrD) + mult*d(1:degrD)
        mult = a(deglN); a=a/mult; b=b/mult
      ELSE
        detB = a(deglN)*d(degrD) - b(degrD)*c(deglN)
        Bmat(1,1) = d(degrD)/detB
        Bmat(1,2) = -c(deglN)/detB
        Bmat(2,1) = -b(degrD)/detB
        Bmat(2,2) = a(deglN)/detB 
        ALLOCATE(ah(degN)); ALLOCATE(ch(degN))
        ALLOCATE(bh(degD)); ALLOCATE(dh(degD))
        ah = a(1:degN); ch = c(1:degN)
        bh = b(1:degD); dh = d(1:degD)
        a(1:degN) = Bmat(1,1)*ah + Bmat(2,1)*ch
        c(1:degN) = Bmat(1,2)*ah + Bmat(2,2)*ch
        b(1:degD) = Bmat(1,1)*bh + Bmat(2,1)*dh
        d(1:degD) = Bmat(1,2)*bh + Bmat(2,2)*dh
        DEALLOCATE(ah); DEALLOCATE(ch)
        DEALLOCATE(bh); DEALLOCATE(dh)
!! new
        a(degN)=1.0_DP;c(degN)=0.0_DP;b(degD)=0.0_DP;d(degD)=1.0_DP
      END IF
     
      capN = m + n

      CALL Powers(Gi,npts,pws)

      CALL dcffti(npts,wsave)

      ALLOCATE(val11(npts)); ALLOCATE(val12(npts))
      ALLOCATE(val21(npts)); ALLOCATE(val22(npts))

      DO k = 1, npts
        val11(k) = pws(k)*a(k)
        val12(k) = pws(k)*c(k)
        val21(k) = pws(k)*b(k)
        val22(k) = pws(k)*d(k)
      END DO 

      val11 = CONJG(val11)
      val12 = CONJG(val12)
      val21 = CONJG(val21)
      val22 = CONJG(val22)

      CALL dcfftf(npts,val11,wsave)
      CALL dcfftf(npts,val12,wsave)
      CALL dcfftf(npts,val21,wsave)
      CALL dcfftf(npts,val22,wsave)

      val11 = CONJG(val11)
      val12 = CONJG(val12)
      val21 = CONJG(val21)
      val22 = CONJG(val22)

      ALLOCATE(dval11(npts)); ALLOCATE(dval12(npts))
      ALLOCATE(dval21(npts)); ALLOCATE(dval22(npts))

      dval11 = 0.0_DP; dval12 = 0.0_DP; dval21 = 0.0_DP; dval22 = 0.0_DP

      DO k = 1, n
        dval11(k) = k*pws(k)*a(k+1)
        dval12(k) = k*pws(k)*c(k+1)
      END DO

      DO k = 1, m
        dval21(k) = k*pws(k)*b(k+1)
        dval22(k) = k*pws(k)*d(k+1)
      END DO

      dval11 = CONJG(dval11)
      dval12 = CONJG(dval12)
      dval21 = CONJG(dval21)
      dval22 = CONJG(dval22)

      CALL dcfftf(npts,dval11,wsave)
      CALL dcfftf(npts,dval12,wsave)
      CALL dcfftf(npts,dval21,wsave)
      CALL dcfftf(npts,dval22,wsave)

      dval11 = CONJG(dval11)
      dval12 = CONJG(dval12)
      dval21 = CONJG(dval21)
      dval22 = CONJG(dval22)

      h1 = 0.0_DP; h2 = 0.0_DP

      DO k = 1, capN
        i = ipoints(k)
        ma = val11(i); mc = val12(i)
        mb = val21(i); md = val22(i)

        IF (ABS(ma)**2 + ABS(mc)**2 < ABS(mb)**2 + ABS(md)**2) THEN
          ma = mb; mc = md
        END IF

        mb = resL(i)*dval11(i) + resR(i)*dval21(i)
        md = resL(i)*dval12(i) + resR(i)*dval22(i)
        detm = ma*md - mb*mc

        h1(i) = -mc/detm
        h2(i) = ma/detm
      END DO

      DEALLOCATE(dval22); DEALLOCATE(dval21)
      DEALLOCATE(dval12); DEALLOCATE(dval11)
      DEALLOCATE(val22); DEALLOCATE(val21)
      DEALLOCATE(val12); DEALLOCATE(val11)

      ALLOCATE(w11e(2*npts)); ALLOCATE(w12e(2*npts)) 
      ALLOCATE(w21e(2*npts)); ALLOCATE(w22e(2*npts)) 

      w11e = 0.0_DP; w12e = 0.0_DP; w21e = 0.0_DP; w22e = 0.0_DP
      
      DO k = 1, n
        w11e(k) = a(n+2-k)
        w12e(k) = c(n+2-k)
      END DO 

      DO k = 1, m
        w21e(k) = b(m+2-k)
        w22e(k) = d(m+2-k)
      END DO

      CALL dcffti(2*npts,wsave2)

      w11e = CONJG(w11e)
      w12e = CONJG(w12e)
      w21e = CONJG(w21e)
      w22e = CONJG(w22e)

      CALL dcfftf(2*npts,w11e,wsave2)
      CALL dcfftf(2*npts,w12e,wsave2)
      CALL dcfftf(2*npts,w21e,wsave2)
      CALL dcfftf(2*npts,w22e,wsave2)

      w11e = CONJG(w11e)
      w12e = CONJG(w12e)
      w21e = CONJG(w21e)
      w22e = CONJG(w22e)

      mn = MAX(m,n)

      mresL = MAXVAL(abs(resL(ipoints(1:capN))));
      mresR = MAXVAL(abs(resR(ipoints(1:capN))));

      IF (verbose > 0) THEN
        print *,'    The maximum value of the left and right residuals is'
        WRITE(*,1012) mresL, mresR
1012    FORMAT(4X,ES24.16,ES24.16)
        print*,' '
        print*,'    The relative error at each iteration step is'
      END IF

      pacc = 10.0E20_DP
      DO j = 0, nitref
        CALL GetRes(a,b,c,d,resL,resR,npts,.TRUE.,Gi,rL,rR)

        mrL = MAXVAL(abs(rL(ipoints(1:capN))));
        mrR = MAXVAL(abs(rR(ipoints(1:capN))));

        IF (verbose > 0) THEN
          WRITE(*,1013) j,mrL/mresL,mrR/mresR
1013      FORMAT(4X,I4,2ES24.16)
        END IF
        
        acc = Max(mrL/mresL,mrR/mresR)

        time02 = GiveTime()

        IF (verbose > 0) THEN
          print*,'    Execution Time = ',time02-time01
        END IF
       
        WRITE(94,8010) j, time02-time01
8010    FORMAT(2X,I5,1X,ES9.2)

        IF ((acc > pacc/factor).OR.(acc .EQ. 0.0_DP)) THEN
          EXIT
        END IF

        pacc = acc

        ALLOCATE(x1(2*npts)); ALLOCATE(x2(2*npts))

        DO k = 1, npts
          x1(k) = -h1(k)*rL(k)
          x2(k) = -h2(k)*rL(k)
        END DO

        CALL dcfftb(npts,x1,wsave)
        CALL dcfftb(npts,x2,wsave)
        x1 = x1/npts
        x2 = x2/npts

        ALLOCATE(x1ae(2*npts)); ALLOCATE(x2ae(2*npts))

        x1ae = 0.0_DP; x2ae = 0.0_DP
   
        DO k = 1, mn
          x1ae(k) = capN * pws(k) * x1(k)
          x2ae(k) = capN * pws(k) * x2(k)
        END DO

        x1ae = CONJG(x1ae)
        x2ae = CONJG(x2ae)

        CALL dcfftf(2*npts,x1ae,wsave2)
        CALL dcfftf(2*npts,x2ae,wsave2)

        x1ae = CONJG(x1ae)
        x2ae = CONJG(x2ae)

        DO k = 1, 2*npts
          x1(k) = w11e(k)*x1ae(k) + w12e(k)*x2ae(k)
          x2(k) = w21e(k)*x1ae(k) + w22e(k)*x2ae(k)
        END DO

        x1 = CONJG(x1)
        x2 = CONJG(x2)

        CALL dcfftb(2*npts,x1,wsave2)
        CALL dcfftb(2*npts,x2,wsave2)
        x1 = x1/(2*npts)
        x2 = x2/(2*npts)

        x1 = CONJG(x1)
        x2 = CONJG(x2)

        a(1:n) = a(1:n) + x1(n:1:-1)
        b(1:m) = b(1:m) + x2(m:1:-1)

        DO k = 1, npts
          x1(k) = -h1(k)*rR(k)
          x2(k) = -h2(k)*rR(k)
        END DO

        CALL dcfftb(npts,x1,wsave)
        CALL dcfftb(npts,x2,wsave)
        x1 = x1/npts
        x2 = x2/npts

        x1ae = 0.0_DP; x2ae = 0.0_DP

        DO k = 1, mn
          x1ae(k) = capN * pws(k) * x1(k)
          x2ae(k) = capN * pws(k) * x2(k)
        END DO

        x1ae = CONJG(x1ae)
        x2ae = CONJG(x2ae)

        CALL dcfftf(2*npts,x1ae,wsave2)
        CALL dcfftf(2*npts,x2ae,wsave2)

        x1ae = CONJG(x1ae)
        x2ae = CONJG(x2ae)

        DO k = 1, 2*npts
          x1(k) = w11e(k)*x1ae(k) + w12e(k)*x2ae(k)
          x2(k) = w21e(k)*x1ae(k) + w22e(k)*x2ae(k)
        END DO

        DEALLOCATE(x2ae); DEALLOCATE(x1ae)

        x1 = CONJG(x1)
        x2 = CONJG(x2)

        CALL dcfftb(2*npts,x1,wsave2)
        CALL dcfftb(2*npts,x2,wsave2)
        x1 = x1/(2*npts)
        x2 = x2/(2*npts)

        x1 = CONJG(x1)
        x2 = CONJG(x2)

        c(1:n) = c(1:n) + x1(n:1:-1)
        d(1:m) = d(1:m) + x2(m:1:-1)

        DEALLOCATE(x1); DEALLOCATE(x2)

      END DO

      DEALLOCATE(w22e); DEALLOCATE(w21e)
      DEALLOCATE(w12e); DEALLOCATE(w11e)

      IF (verbose > 0) THEN
        print*,' '
        print*,'   *** END ITERATIVE REFINEMENT ***'
        print*,' '
      END IF

      END SUBROUTINE ItRefSf


!-----------------------------------------------------------------------


      SUBROUTINE Extract(n,h,ZZ,FF)
!-----------------------------------------------------------------------
!**PURPOSE
!
!  Extracts the Loewner information from the Hankel data.
!
!-----------------------------------------------------------------------
!  Parameters
!
! h DIMENSION(0:2*n-2)
! ZZ, FF DIMENSION(2*n)
      INTEGER, INTENT(IN)                              :: n
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN) :: h
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(OUT)    :: ZZ, FF

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                                     :: k
      INTEGER, DIMENSION(2*n)                     :: pivot, invpivot
      REAL(KIND=DP)                               :: pi
      REAL(KIND=DP), DIMENSION(4*n+15)            :: wsave
      REAL(KIND=DP), DIMENSION(8*n+15)            :: wsave2
      COMPLEX(KIND=DP)                            :: zeta
      COMPLEX(KIND=DP), PARAMETER                 :: I = (0.,1.)
      COMPLEX(KIND=DP), DIMENSION(n)              :: a, y, z, zeta_n
      COMPLEX(KIND=DP), DIMENSION(n)              :: c, d
      COMPLEX(KIND=DP), DIMENSION(2*n)            :: a2, zeta_2n, hvec
      COMPLEX(KIND=DP), DIMENSION(2*n)            :: param

!-----------------------------------------------------------------------


      pi = 4.0_DP*ATAN(1.0_DP)

      CALL dcffti(n,wsave)
      a = 0.0_DP
      a(2) = 1.0_DP
      CALL dcfftf(n,a,wsave)
      y = CONJG(a)
      zeta = EXP(pi*I/n)
      z = zeta*y

      CALL dcffti(2*n,wsave2)
      a2 = 0.0_DP
      a2(2) = 1.0_DP
      CALL dcfftf(2*n,a2,wsave2)
      zeta_2n = CONJG(a2)
      zeta_n = zeta_2n(1:n)

      DO k = 1, n
        pivot(2*k-1) = k
        pivot(2*k) = n+k
        invpivot(k) = 2*k-2
        invpivot(k+n) = 2*k-1
      END DO
      invpivot = 1 + invpivot

      hvec(1:2*n-1) = h(0:2*n-2)
      hvec(2*n) = 0.0_DP
      CALL dcfftf(2*n,hvec,wsave2)
      param = CONJG(zeta_2n) * hvec
      param = param(invpivot)
      c = param(1:n)
      d = param(n+1:2*n)

      ZZ(1:n) = y; ZZ(n+1:2*n) = z
      FF(1:n) = c; FF(n+1:2*n) = d
      ZZ = ZZ(pivot)
      FF = FF(pivot)

      END SUBROUTINE Extract


!-----------------------------------------------------------------------


      SUBROUTINE GetRes(a,b,c,d,L,R,n,show,Gi,resL,resR) 
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
! a,b,c,d,L,R, resL, resR DIMENSION(n)
! "show" is not used but can be used for test purposes
      INTEGER, INTENT(IN)                          :: n, Gi
      LOGICAL, INTENT(IN)                          :: show
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)   :: a,b,c,d,L,R
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(OUT)  :: resL, resR

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                            :: k
      REAL(KIND=DP), DIMENSION(4*n+15)   :: wsave
      COMPLEX(KIND=DP), DIMENSION(n)     :: aeval, beval, ceval, deval
      COMPLEX(KIND=DP), DIMENSION(n)     :: pws

!-----------------------------------------------------------------------

      aeval = a; beval = b; ceval = c; deval = d

      CALL Powers(Gi,n,pws)
      CALL dcffti(n,wsave)

      IF (Size(a,1) > n) THEN
        resL = 0.0_DP
      ELSE
        DO k = 1, n
          aeval(k) = pws(k)*aeval(k)
          beval(k) = pws(k)*beval(k)
        END DO
        aeval = CONJG(aeval)
        beval = CONJG(beval)
        CALL dcfftf(n,aeval,wsave)
        CALL dcfftf(n,beval,wsave)
        aeval = CONJG(aeval)
        beval = CONJG(beval)
        DO k = 1, n
          resL(k) = L(k)*aeval(k) + R(k)*beval(k)
        END DO
      END IF

      IF (Size(d,1) > n) THEN
        resR = 0.0_DP
      ELSE
        DO k = 1, n
          ceval(k) = pws(k)*ceval(k)
          deval(k) = pws(k)*deval(k)
        END DO
        ceval = CONJG(ceval)
        deval = CONJG(deval)
        CALL dcfftf(n,ceval,wsave)
        CALL dcfftf(n,deval,wsave)
        ceval = CONJG(ceval)
        deval = CONJG(deval)
        DO k = 1, n
          resR(k) = L(k)*ceval(k) + R(k)*deval(k)
        END DO
      END IF

      END SUBROUTINE GetRes      


!-----------------------------------------------------------------------


      SUBROUTINE PlotRes(n,resL,resR,filenum,filename,itype)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
! resL, resR DIMENSION(n)
! "filename" can be used for test purposes
      INTEGER, INTENT(IN)               :: n, filenum, itype
      CHARACTER (LEN=*), INTENT(IN)     :: filename
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)  :: resL, resR

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER  k
      REAL(KIND=DP), DIMENSION(n):: residual

!-----------------------------------------------------------------------

!!      OPEN(filenum,FILE=filename,STATUS='REPLACE')

      IF (itype == 1) THEN
        residual = ABS(resL)
        WRITE(filenum,1010) n,SUM(residual)/n,MINVAL(residual),MAXVAL(residual)
        residual = ABS(resR)
        WRITE(filenum,1010) n,SUM(residual)/n,MINVAL(residual),MAXVAL(residual)

      ELSE IF (itype == 2) THEN
        DO k = 1, n
!          WRITE(filenum,100) LOG10(ABS(resL(k))), LOG10(ABS(resR(k)))
          WRITE(filenum,101) (resL(k)), (resR(k))
        END DO
      END IF


!!     CLOSE(filenum)

100   FORMAT(F6.2,' ',F6.2,' ',F6.2,' ',F6.2)
101   FORMAT(4E10.2)
1010  FORMAT(I8,E10.2,E10.2,E10.2)

      END SUBROUTINE PlotRes


!-----------------------------------------------------------------------


      SUBROUTINE Powers(Gi,n,pws)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
! pws DIMENSION(n)
      INTEGER, INTENT(IN)                         :: Gi, n
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(OUT) :: pws

!-----------------------------------------------------------------------
!  Local variables
!
!
      INTEGER, DIMENSION(n) :: ii
      INTEGER i

!-----------------------------------------------------------------------
      IF (Gi == 0) THEN
        pws = 1.0_DP
      ELSE
!        ii = (/ (i,i=0,(n-1)*Gi,Gi) /)
!        ii = MOD(ii,Gn) + 1
        ii(1) = 0
        DO i =2, n
          ii(i) = MOD(ii(i-1)+Gi,Gn)
        END DO
        ii = ii + 1
        pws = Gz(ii)
      END IF
      
      END SUBROUTINE Powers


!-----------------------------------------------------------------------


      SUBROUTINE LRScaling(resL,resR,alpha)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT) :: resL, resR
      REAL(KIND=DP), INTENT(OUT) :: alpha

!-----------------------------------------------------------------------
!  Local variables
!
      REAL(KIND=DP) alphaL, alphaR

!-----------------------------------------------------------------------
      alphaL = MAXVAL(ABS(resL));
      alphaR = MAXVAL(ABS(resR));

      resL = resL/alphaL
      resR = resR/alphaR

      alpha = alphaR/alphaL

      END SUBROUTINE LRScaling


!-----------------------------------------------------------------------


      SUBROUTINE GScaling(resL,resR)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT) :: resL, resR

!-----------------------------------------------------------------------
!  Local variables
!
      REAL(KIND=DP) :: alpha, alphaL, alphaR

!-----------------------------------------------------------------------
      alphaL = MAXVAL(ABS(resL));
      alphaR = MAXVAL(ABS(resR));
      alpha = MAX(alphaL,alphaR)

      resL = resL/alpha
      resR = resR/alpha

      END SUBROUTINE GScaling


!-----------------------------------------------------------------------


      SUBROUTINE PScaling(resL,resR)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT) :: resL, resR

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER       :: n,i
      REAL(KIND=DP) :: alpha

!-----------------------------------------------------------------------
       n = Size(resL)
       DO i = 1, n
         alpha = MAX(ABS(resL(i)), ABS(resR(i)))
         resL(i) = resL(i)/alpha
         resR(i) = resR(i)/alpha
       END DO

      END SUBROUTINE PScaling


!-----------------------------------------------------------------------


      SUBROUTINE V2Max(a,b,maximum,where)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
      REAL(KIND=DP), DIMENSION(:), INTENT(IN) :: a
      REAL(KIND=DP), DIMENSION(SIZE(a)), INTENT(IN) :: b
      REAL(KIND=DP), INTENT(OUT) :: maximum
      INTEGER, INTENT(OUT) :: where 

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER, DIMENSION(1) :: subscript

!-----------------------------------------------------------------------
       subscript = MAXLOC(MAX(a,b))
       where = subscript(1)
       maximum = MAX(a(where),b(where))


      END SUBROUTINE V2Max


!-----------------------------------------------------------------------


      SUBROUTINE PolInterp(c,n,Gi)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
! c DIMENSION(n)
      INTEGER, INTENT(IN)                           :: n, Gi
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT) :: c

!-----------------------------------------------------------------------
!  Local variables
!
      REAL(KIND=DP), DIMENSION(4*n+15)         :: wsave
      COMPLEX(KIND=DP), DIMENSION(n)           :: pws

!-----------------------------------------------------------------------
        CALL dcffti(n,wsave)
        c = Conjg(c)
        CALL dcfftb(n,c,wsave)
        c = Conjg(c)
        CALL Powers(Gi,n,pws)
        c = c / pws
        c = c / n

      END SUBROUTINE PolInterp


!-----------------------------------------------------------------------


      SUBROUTINE PolMatMult(a1,b1,c1,d1,a2,b2,c2,d2,state1,state2,n,a,b,c,d)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
! a1, b1, c1, d1, a2, b2, c2, d2 DIMENSION(n/2)
! a, b, c, d DIMENSION(n)
      INTEGER, INTENT(IN)                          :: n
      INTEGER, DIMENSION(3), INTENT(IN)            :: state1, state2
      COMPLEX(KIND=DP), DIMENSION(:),   INTENT(OUT):: a, b, c, d
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)   :: a1, b1, c1, d1
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)   :: a2, b2, c2, d2

!-----------------------------------------------------------------------
!  Local variables
!
!
      INTEGER                          :: degl1, degr1, tau1
      INTEGER                          :: degl2, degr2, tau2
      INTEGER                          :: tau, ndiv2
      REAL(KIND=DP), DIMENSION(4*n+15) :: wsave
      COMPLEX(KIND=DP), DIMENSION(n)   :: a1eval, b1eval
      COMPLEX(KIND=DP), DIMENSION(n)   :: c1eval, d1eval
      COMPLEX(KIND=DP), DIMENSION(n)   :: a2eval, b2eval
      COMPLEX(KIND=DP), DIMENSION(n)   :: c2eval, d2eval
 
!-----------------------------------------------------------------------
        ndiv2 = n/2
        tau1 = state1(1); degl1 = state1(2); degr1 = state1(3)
        tau2 = state2(1); degl2 = state2(2); degr2 = state2(3)

        tau = tau1

        a1eval = 0.0_DP; a1eval(1:ndiv2) = a1
        IF (degl1 > ndiv2) THEN
          a1eval(ndiv2+1) = 1.0_DP
        END IF
        b1eval = 0.0_DP; b1eval(1:ndiv2) = b1
        c1eval = 0.0_DP; c1eval(1:ndiv2) = c1
        d1eval = 0.0_DP; d1eval(1:ndiv2) = d1
        IF (degr1+tau1 > ndiv2) THEN
          d1eval(ndiv2+1) = 1.0_DP
        END IF

        a1eval = CONJG(a1eval)
        b1eval = CONJG(b1eval)
        c1eval = CONJG(c1eval)
        d1eval = CONJG(d1eval)

        CALL dcffti(n,wsave)

        CALL dcfftf(n,a1eval,wsave)
        CALL dcfftf(n,b1eval,wsave)
        CALL dcfftf(n,c1eval,wsave)
        CALL dcfftf(n,d1eval,wsave)

        a1eval = CONJG(a1eval)
        b1eval = CONJG(b1eval)
        c1eval = CONJG(c1eval)
        d1eval = CONJG(d1eval)

        a2eval = 0.0_DP; a2eval(1:ndiv2) = a2
        IF (degl2 > ndiv2) THEN
          a2eval(ndiv2+1) = 1.0_DP
        END IF
        b2eval = 0.0_DP; b2eval(1:ndiv2) = b2
        c2eval = 0.0_DP; c2eval(1:ndiv2) = c2
        d2eval = 0.0_DP; d2eval(1:ndiv2) = d2
        IF (degr2+tau2 > ndiv2) THEN
          d2eval(ndiv2+1) = 1.0_DP
        END IF

        a2eval = CONJG(a2eval)
        b2eval = CONJG(b2eval)
        c2eval = CONJG(c2eval)
        d2eval = CONJG(d2eval)
 
        CALL dcfftf(n,a2eval,wsave)
        CALL dcfftf(n,b2eval,wsave)
        CALL dcfftf(n,c2eval,wsave)
        CALL dcfftf(n,d2eval,wsave)


        a2eval = CONJG(a2eval)
        b2eval = CONJG(b2eval)
        c2eval = CONJG(c2eval)
        d2eval = CONJG(d2eval)

        a = a1eval * a2eval + c1eval * b2eval
        b = b1eval * a2eval + d1eval * b2eval
        c = a1eval * c2eval + c1eval * d2eval
        d = b1eval * c2eval + d1eval * d2eval

        a = CONJG(a)
        b = CONJG(b)
        c = CONJG(c)
        d = CONJG(d)

        CALL dcfftb(n,a,wsave) 
        CALL dcfftb(n,b,wsave) 
        CALL dcfftb(n,c,wsave) 
        CALL dcfftb(n,d,wsave) 

        a = a/n
        b = b/n
        c = c/n
        d = d/n

        a = CONJG(a)
        b = CONJG(b)
        c = CONJG(c)
        d = CONJG(d)

      END SUBROUTINE PolMatMult


!-----------------------------------------------------------------------


      SUBROUTINE Downdating(a,b,c,d,points, &
                            L,R,n,show,Gi,state,ipoints,ndowndating) 
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
! a,b,c,d, L,R, resL, resR DIMENSION(n)
      INTEGER, INTENT(IN)                        :: n, Gi
      INTEGER, DIMENSION(3), INTENT(INOUT)       :: state
      INTEGER, DIMENSION(n), INTENT(INOUT)       :: ipoints
      INTEGER, INTENT(OUT)                       :: ndowndating
      INTEGER, INTENT(IN)                        :: show
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(INOUT)   :: a,b,c,d
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)   :: L,R,points

!-----------------------------------------------------------------------
!  Local variables
!
      LOGICAL                          :: changeleft, combination
      INTEGER                          :: k, capN, tau, degl, degr
      INTEGER                          :: deglN, degrN, deglD, degrD
      INTEGER                          :: finalk, iaux, kk, nn, hn
      INTEGER, DIMENSION(1)            :: ipointsaux
      REAL(KIND=DP)                 :: wsave(4*n+15)
      REAL(KIND=DP)                 :: unorm, lnorm
      REAL(KIND=DP)                 :: maxdetB
      COMPLEX(KIND=DP), DIMENSION(:), allocatable  :: resL, resR
      COMPLEX(KIND=DP), DIMENSION(:), allocatable  :: resLcheck, resRcheck
      COMPLEX(KIND=DP), DIMENSION(n)     :: aeval, beval, ceval, deval
      COMPLEX(KIND=DP), DIMENSION(n)     :: pws
      COMPLEX(KIND=DP)                   :: vl, vr
      COMPLEX(KIND=DP)                   :: alpha
      COMPLEX(KIND=DP), DIMENSION(n)     :: detB
      COMPLEX(KIND=DP), DIMENSION(1)     :: resLnew, resRnew, point
      COMPLEX(KIND=DP)                   :: aevalkk, bevalkk, cevalkk, devalkk
      
!-----------------------------------------------------------------------


      tau = state(1); degl = state(2); degr = state(3)

      deglN = degl; deglD = degl + tau; degrN = degr; degrD = degr + tau

      aeval = a; beval = b; ceval = c; deval = d

      CALL Powers(Gi,n,pws)
      CALL dcffti(n,wsave)

      DO k = 1, n
        aeval(k) = pws(k)*aeval(k)
        beval(k) = pws(k)*beval(k)
      END DO
      aeval = CONJG(aeval)
      beval = CONJG(beval)
      CALL dcfftf(n,aeval,wsave)
      CALL dcfftf(n,beval,wsave)
      aeval = CONJG(aeval)
      beval = CONJG(beval)
      ALLOCATE(resL(n))
      ALLOCATE(resLcheck(n))
      DO k = 1, n
        resL(k) = L(k)*aeval(k) + R(k)*beval(k)
      END DO

      DO k = 1, n
        ceval(k) = pws(k)*ceval(k)
        deval(k) = pws(k)*deval(k)
      END DO
      ceval = CONJG(ceval)
      deval = CONJG(deval)
      CALL dcfftf(n,ceval,wsave)
      CALL dcfftf(n,deval,wsave)
      ceval = CONJG(ceval)
      deval = CONJG(deval)
      ALLOCATE(resR(n))
      ALLOCATE(resRcheck(n))
      DO k = 1, n
        resR(k) = L(k)*ceval(k) + R(k)*deval(k)
      END DO

      IF (show > 0) CALL PlotRes(n,resL,resR,5,'plotres.dat',show)

      capN = degl + degr + tau - 2
print*,'MAXVAL(ABS(L)) =    ',MAXVAL(ABS(L(ipoints(1:capN))))
print*,'MAXVAL(ABS(R)) =    ',MAXVAL(ABS(R(ipoints(1:capN))))
print*,'MAXVAL(ABS(resL)) = ',MAXVAL(ABS(resL(ipoints(1:capN))))
print*,'MAXVAL(ABS(resR)) = ',MAXVAL(ABS(resR(ipoints(1:capN))))

      DO k = 1, n
        detB(k) = aeval(k)*deval(k)-beval(k)*ceval(k)
      END DO

      maxdetB = MAXVAL(ABS(detB(ipoints(1:capN))))
print*,'maxdetB =',maxdetB

      ndowndating = 0

      finalk = capN

      k = 1

      DO WHILE (k <= finalk)

        IF ( ABS(resL(ipoints(k))) > Gthreshold .OR.     &
             ABS(resR(ipoints(k))) > Gthreshold) THEN

          iaux = ipoints(k)
          ipoints(k) = ipoints(finalk)
          ipoints(finalk) = iaux

          finalk = finalk - 1

        ELSE

          k = k + 1

        ENDIF

      END DO

      ndowndating = capN - finalk

      k = finalk + 1
      nn = finalk
      finalk = capN

      DO WHILE (k <= finalk)

        kk = ipoints(k)

        aevalkk = PolVal(a,points(kk))
        bevalkk = PolVal(b,points(kk))
        cevalkk = PolVal(c,points(kk))
        devalkk = PolVal(d,points(kk))

        unorm = SQRT ( ABS(aevalkk)**2 + ABS(cevalkk)**2 )
        lnorm = SQRT ( ABS(bevalkk)**2 + ABS(devalkk)**2 )

        IF (unorm > lnorm) THEN
          vl = -cevalkk/unorm
          vr =  aevalkk/unorm
        ELSE
          vl =  devalkk/lnorm
          vr = -bevalkk/lnorm
        ENDIF

        IF (degl == degr) THEN

          combination = .TRUE.
          changeleft = abs(vl) > abs(vr)

        ELSEIF (degl > degr) THEN

          combination = abs(vl) > Gnumzero
          changeleft = combination

        ELSE

          combination = abs(vr) > Gnumzero
          changeleft = .NOT. combination

        ENDIF
      
        IF ( combination ) THEN

          IF (verbose > 0) THEN
            print*,'    Downdating ', k
          END IF
          
          IF ( .NOT. changeleft ) THEN

            alpha = vl/vr

            c = alpha * a + c
            d = alpha * b + d

          ELSE

            alpha = vr/vl
print*,'alpha = ',alpha
print*,'aevalkk=',aevalkk+alpha*cevalkk
print*,'bevalkk=',bevalkk+alpha*devalkk

            a = a + alpha * c
            b = b + alpha * d

          ENDIF

        ENDIF

        IF (.NOT. changeleft) THEN

! c, d, ceval, deval and resR : divide by (z-z_k)

            CALL DivideCoeff(c,points(kk))
            CALL DivideCoeff(d,points(kk))

            degr = degr - 1

        ELSE

! a, b, aeval, beval and resL : divide by (z-z_k)

            CALL DivideCoeff(a,points(kk))
            CALL DivideCoeff(b,points(kk))

            degl = degl - 1

        ENDIF

!        state(2) = degl; state(3) = degr
!
!        resLnew = L(kk)*PolVal(a,points(kk))+R(kk)*PolVal(b,points(kk))
!        resRnew = L(kk)*PolVal(c,points(kk))+R(kk)*PolVal(d,points(kk))
!
!        point   = points(kk:kk)
!
!        CALL RatInt(point,resLnew,resRnew, &
!                        1,Gfactor,state, &
!                        a,c,b,d,hn,ipointsaux,0,0,.TRUE.)
!
!        degl = state(2); degr = state(3)
!
        k = k + 1

      END DO

      state(2) = degl; state(3) = degr

      IF (show > 0) THEN
        CALL GetRes(a,b,c,d,L,R,n,.TRUE.,Gi,resL,resR)
        CALL PlotRes(n,resL,resR,5,'plotres.dat',show)
      END IF

      IF (verbose > 0) THEN
        print*,'    number of interpolation conditions downdated = ',&
               ndowndating
      END IF

!!??
!      ndowndating = 0
!!??

      DEALLOCATE(resL); DEALLOCATE(resR)
      DEALLOCATE(resLcheck); DEALLOCATE(resRcheck)

      END SUBROUTINE Downdating      


!-----------------------------------------------------------------------


      SUBROUTINE DivideVal(v,k)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
! v DIMENSION(Gn)
      INTEGER,                        INTENT(in)     :: k
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(inout)  :: v

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                          :: i
      COMPLEX(KIND=DP)                 :: Gzk

!-----------------------------------------------------------------------
      Gzk = Gz(k)

      DO i = 1, k-1
        v(i) = v(i) / (Gz(i)-Gzk)
      END DO

      DO i = k+1, Gn
        v(i) = v(i) / (Gz(i)-Gzk)
      END DO

      END SUBROUTINE DivideVal


!-----------------------------------------------------------------------


      SUBROUTINE DivideCoeff(coeff,point)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
      COMPLEX(KIND=DP),               INTENT(in)     :: point
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(inout)  :: coeff

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                          :: i, ncoeff

!-----------------------------------------------------------------------

      ncoeff = Size(coeff)

      DO i = ncoeff-1,1,-1
        coeff(i) = coeff(i) + point*coeff(i+1)
      END DO

print*,'Remainder in DivideCoeff =',coeff(1)

      DO i = 1, ncoeff-1
        coeff(i) = coeff(i+1)
      END DO

      coeff(ncoeff) = 0.0_DP

      END SUBROUTINE DivideCoeff


!-----------------------------------------------------------------------


      COMPLEX(KIND=DP) FUNCTION PolVal(coeff,point)
!-----------------------------------------------------------------------
!**PURPOSE
!!
!-----------------------------------------------------------------------
!  Parameters
!
      COMPLEX(KIND=DP),            INTENT(in)     :: point
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(in)  :: coeff

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                          :: i, ncoeff
      COMPLEX(KIND=DP), DIMENSION(Size(coeff))  :: quotient

!-----------------------------------------------------------------------
      ncoeff = Size(coeff)
print*,'ncoeff in PolVal = ',ncoeff

      quotient(ncoeff) = 0.0_DP
      quotient(ncoeff-1) = coeff(ncoeff)
      DO i = ncoeff-2,1,-1
        quotient(i) = coeff(i+1) + point*quotient(i+1)
      END DO

      PolVal = coeff(1) + point*quotient(1)

      RETURN

      END FUNCTION PolVal


!-----------------------------------------------------------------------

END MODULE RatInt_Module
