!-----------------------------------------------------------------------
!**BEGIN PROLOGUE Toeplitz_Module
!**DATE WRITTEN   1998/10/01  (year/month/day)
!**AUTHORS
!
!  Marc Van Barel, Peter Kravanja
!
!    Department of Computer Science
!    Katholieke Universiteit Leuven
!    Celestijnenlaan 200 A
!    B-3001 Heverlee
!    BELGIUM
!    E-mail: Marc.VanBarel@cs.kuleuven.ac.be
!            Peter.Kravanja@na-net.ornl.gov
!
!**PURPOSE
!
! This module contains subroutines for transforming the Toeplitz problem
! Tx = b into the associated rational interpolation problem and back.
! It also contains functions for the efficient multiplication of a Toeplitz
! matrix and a column vector, and for evaluating the product of the inverse
! of a Toeplitz matrix with a column vector, given the parameters that appear
! in the inversion formula for the Toeplitz matrix.
!
!**END PROLOGUE Toeplitz_Module
!-----------------------------------------------------------------------

MODULE Toeplitz_Module

     USE Precision
     USE Timing
     USE Verbose_Module
     USE RatInt_Module

     IMPLICIT NONE
!-----------------------------------------------------------------------
!**ACCESSABILITY
!
     PRIVATE
     PUBLIC :: Transform_T_RIP, Transform_RIP_T, Tx, Tx2, InvToeplitz


CONTAINS


      SUBROUTINE Transform_T_RIP(n,a,b)
!-----------------------------------------------------------------------
!**PURPOSE
!
!   Input Parameters
!   ----------------
!
!   n       Integer
!           The size of the system.
!
!   a       Double complex array of dimension (1:2*n-1)
!
!   b       Double complex array of dimension (1:n)
!           The right hand side.
!
!-----------------------------------------------------------------------
!  Parameters
!
      INTEGER, INTENT(IN)                            :: n
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)     :: a, b

!-----------------------------------------------------------------------
!  Local variables
!
      INTEGER                             :: N2, capN, Nmn, Nmn2, tau, ii
      INTEGER, DIMENSION(:), ALLOCATABLE  :: index
      REAL(KIND=DP), DIMENSION(:), ALLOCATABLE  :: wsave
      REAL(KIND=DP)                             :: time01, time02

!-----------------------------------------------------------------------
!  External routines

      EXTERNAL dcffti, dcfftf, dcfftb

!-----------------------------------------------------------------------

        time01 = GiveTime()
        time01 = GiveTime()

        capN = 1
        DO WHILE (capN < n)
          capN = capN *2
        END DO

        Nmn = capN - n
        N2 = 2*capN

        ALLOCATE(Gz(N2))
        Gn = N2
        Gz = 0.0_DP
        Gz(2) = 1.0_DP
        ALLOCATE(wsave(4*N2+15))
        CALL dcffti(N2,wsave)
        CALL dcfftf(N2,Gz,wsave)
        Gz = Conjg(Gz)

        ALLOCATE(index(N2))

        Nmn2 = 2 * Nmn
        index(1) = 0
        DO ii=1,N2-1
          index(ii+1) = MOD(index(ii)+Nmn2,N2)
        END DO
        index = index + 1
        
        ALLOCATE(resR(N2))
        resR = 0.0_DP
        resR(2:2*n) = Conjg(a(1:2*n-1))
        CALL dcfftf(N2,resR,wsave)
        DEALLOCATE(wsave)
        resR = Gz(index) * Conjg(resR)
        resR = - resR
        DEALLOCATE(index)
        ALLOCATE(resL(N2))
        resL = 1.0_DP

        tau = -2*Nmn

        time02 = GiveTime()

        IF (verbose > 0) THEN
          print*,'  Execution time without file handling = ',time02-time01
        END IF

!!        WRITE(94,8010) time02-time01
8010    FORMAT(ES9.2)

        WRITE (11) verbose
        WRITE (11) N2
        WRITE (11) Gz
        WRITE (11) resL,resR
        WRITE (11) tau

       RETURN

      END SUBROUTINE Transform_T_RIP


!-----------------------------------------------------------------------


      SUBROUTINE Transform_RIP_T(n,a,b,x,nref,factor)
!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
!
!   Input Parameters
!   ----------------
!
!   n       Integer
!           The size of the system.
!
!   a       Double complex array of dimension (1:2*n-1)
!
!   b       Double complex array of dimension (1:n)
!           The right hand side.
!
!   nref    Integer, optional (default nref = 1)
!
!           The number of times that the computed solution is to be
!           improved iteratively. Each step costs O(n log n) flops.
!
!   factor  Double precision, optional (default factor = 2D+00)
!           Parameter of the stopping criterion of iterative refinement.
!
!   Output Parameters
!   -----------------
!
!   x       Double complex array of dimension (1:n) 
!           The computed solution to Tx=b.
!
!
!***END PROLOGUE Transform_RIP_T
!-----------------------------------------------------------------------
!
!  Parameters
!
! b, x DIMENSION(n)
! a DIMENSION(1:2*n-1)
      INTEGER, INTENT(IN)                            :: n
      INTEGER, OPTIONAL, INTENT(IN)                  :: nref 
      REAL(KIND=DP), OPTIONAL, INTENT(IN)            :: factor
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN)     :: a, b
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(OUT)    :: x
!-----------------------------------------------------------------------
!
!  Local variables
!
      COMPLEX(KIND=DP)                         :: b11, b12, b21, b22
      COMPLEX(KIND=DP)                         :: det, a11, a12, a21, a22
      INTEGER                                  :: N2, capN, j, k
      REAL(KIND=DP), DIMENSION(:), ALLOCATABLE :: R
      LOGICAL                                  :: a_is_zero, test
      INTEGER                                  :: numref, hn
      REAL(KIND=DP)                            :: numfactor, alpha
      REAL(KIND=DP), DIMENSION(:), ALLOCATABLE :: wsave
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE :: res, xr, bx
      REAL(KIND=DP)                               :: pacc, acc, maxb, maxr
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE :: dl, dlh, dr
      REAL(KIND=DP) :: time01, time02
!!
  REAL(KIND=DP)                            :: sumb
!-----------------------------------------------------------------------
!
!  External routines
!
      EXTERNAL dcffti, dcfftf, dcfftb
!-----------------------------------------------------------------------

        IF (PRESENT(nref)) THEN
          numref = nref
        ELSE
          numref = 1
        END IF 
  
        IF (PRESENT(factor)) THEN
          numfactor = ABS(factor)
        ELSE
          numfactor = 2.0E+00_DP
        END IF
  
        a_is_zero = .TRUE.
        DO k = 1, 2*n-1
          test = (a(k) == 0.0_DP)
          a_is_zero = a_is_zero .AND. test
        END DO

!!      a_is_zero = (a == (/ (0.0_DP, k = 1, 2*n-1) /) )

        IF (a_is_zero) THEN
          IF (verbose > 0) THEN
            print*,'  The Toeplitz-matrix is the zero matrix.'
          END IF
          STOP
        END IF

        IF (n == 1) THEN 
            x(1) = b(1)/a(1)
        ELSE

          READ(12) verbose
          READ(12) N2
          ALLOCATE(dl(N2))
          ALLOCATE(dr(N2))
          READ(12) dl
          READ(12) dr
          READ(12) alpha
          READ(12) hn

4001      FORMAT(I10)

          IF (hn < N2) THEN
            IF (verbose > 0) THEN
              print*,'  Not all interpolation conditions are handled.'
            END IF
            STOP
          ELSE

           ALLOCATE(R(n+1))

           R(1:n) = a(n:1:-1); R(n+1) = 0.0_DP
           a11 = DOT_PRODUCT(R,dl(1:n+1))
           a12 = DOT_PRODUCT(R,dr(1:n+1))

           DEALLOCATE(R)
           a21 = dl(n+1); a22 = dr(n+1)
           det = a11*a22 - a12*a21
           b11 = a22/det; b12 = -a12/det
           b21 = -a21/det; b22 = a11/det
           ALLOCATE(dlh(N2))
           dlh = b11*dl + b21*dr
           dr  = dl*b12 + dr*b22
           dl = dlh
           DEALLOCATE(dlh)
           dl = Conjg(dl)
           dr = Conjg(dr)
           ALLOCATE(wsave(4*N2+15))
           CALL dcffti(N2,wsave)
           CALL dcfftf(N2,dl,wsave)
           CALL dcfftf(N2,dr,wsave)
           dl = Conjg(dl) ! v
           dr = Conjg(dr) ! u
           capN = N2 / 2
!!    Gz could also be read in from file "inratint.dat" !!
!!    Gz is needed in subroutine InvToeplitz
           ALLOCATE(Gz(N2))
           Gn = N2
           Gz = 0.0_DP
           Gz(2) = 1.0_DP
           CALL dcfftf(N2,Gz,wsave)
           DEALLOCATE(wsave)
           Gz = Conjg(Gz)

           x = InvToeplitz(dr,dl,b,n,capN)
           IF (verbose > 0) THEN
             print*,'  The first ',MIN(10,n),' components of the computed ',&
                    'solution'
             print*,'  before iterative refinement are:'
             WRITE(*,1068) x(1:MIN(10,n))
1068         FORMAT(2X,2ES24.16)
           END IF
!!
  numref = 10
!!  OPEN(96,FILE='itreferror.txt',STATUS='OLD',POSITION='APPEND')
  OPEN(96,FILE='itreferror.txt',STATUS='REPLACE')
  sumb = SUM(ABS(b))
!!

           IF ( numref > 0) THEN
            time01 = GiveTime()
            pacc = 10.0E+30_DP
            maxb = MAXVAL(abs(b))
            ALLOCATE(res(n))
            ALLOCATE(xr(n))
            ALLOCATE(bx(n))
            IF (verbose > 0) THEN
              print*,' '
              print*,'  *** BEGIN ITERATIVE REFINEMENT ***'
              print*,' '
              print*,'    The inf-norm of the residual vector relative to'
              print*,'    the inf-norm of the right-hand side is:'
            END IF
            DO j = 1, numref
              bx = Tx(a,x,N2)
              res = b - bx
              maxr = MAXVAL(abs(res))
              acc = maxr/maxb
!!
  WRITE(96,1070) j-1,SUM(ABS(res))/sumb,SUM(ABS(x-1.0_DP))/n
1070            FORMAT (4X,I4,2ES24.16)
!!

              time02 = GiveTime()
              IF (verbose > 0) THEN
                WRITE(*,1069) j-1,acc
1069            FORMAT (4X,I4,ES24.16)
                print*,'    Execution Time = ',time02-time01
              END IF

              WRITE(94,8010) j-1, time02-time01
8010          FORMAT(I5,2X,ES9.2)

              IF ((acc > pacc/numfactor).OR.(acc .EQ. 0.0_DP)) THEN
!!                EXIT
              END IF

              pacc = acc

              xr = InvToeplitz(dr,dl,res,n,capN)

              x = x + xr

4002          FORMAT(I10,I10)
            END DO

            time02 = GiveTime()
            WRITE(94,8010) j-1, time02-time01

            DEALLOCATE(res)
            DEALLOCATE(xr)
            DEALLOCATE(bx)

            IF (verbose > 0) THEN
              print*,' '
              print*,'  *** END ITERATIVE REFINEMENT ***'
              print*,' '
            END IF

           END IF

           DEALLOCATE(Gz)
   
          END IF
     
        END IF

      END SUBROUTINE Transform_RIP_T


!-----------------------------------------------------------------------

      FUNCTION Tx2(a,x,N2)
      INTEGER, INTENT(IN) :: N2
!      COMPLEX(KIND=DP), DIMENSION(2*Size(x)-1), INTENT(IN) :: a
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN) :: a
!      COMPLEX(KIND=DP), DIMENSION(n) :: C, R
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN) :: x
      COMPLEX(KIND=DP), DIMENSION(SIZE(x)) :: Tx2
!      COMPLEX(KIND=DP), DIMENSION(N2) :: fftx, fftxflip, fftC, fftR
!      COMPLEX(KIND=DP), DIMENSION(N2) :: Cxeval, Rxeval
      COMPLEX(KIND=DP), DIMENSION(:), ALLOCATABLE :: aux1, aux2
      REAL(KIND=DP), DIMENSION(:), ALLOCATABLE :: wsave
      INTEGER :: n
 
      EXTERNAL dcffti, dcfftf, dcfftb

ALLOCATE(aux1(N2))
ALLOCATE(aux2(N2))
ALLOCATE(wsave(4*N2+15))

      n = Size(x)
      CALL dcffti(N2,wsave)
!      fftx = 0.0_DP
aux1 = 0.0_DP
aux1(1:n) = x
CALL dcfftf(N2,aux1,wsave)

aux2 = 0.0_DP
aux2(1:n) = a(n:2*n-1)
CALL dcfftf(N2,aux2,wsave)

!aux1 = HP(aux1,aux2)
aux1 = aux1 * aux2
CALL dcfftb(N2,aux1,wsave)
Tx2 = aux1(1:n)

aux1 = 0.0_DP
aux1(1:n) = x(n:1:-1)
CALL dcfftf(N2,aux1,wsave)
aux2 = 0.0_DP
aux2(1:n) = a(n:1:-1)
CALL dcfftf(N2,aux2,wsave)
!aux1 = HP(aux1,aux2)
aux1 = aux1 * aux2
CALL dcfftb(N2,aux1,wsave)
Tx2 = (Tx2 + aux1(n:1:-1))/N2

DEALLOCATE(aux1)
DEALLOCATE(aux2)
DEALLOCATE(wsave)

      END FUNCTION Tx2

!----------------------------------------------------------
!
!-----------------------------------------------------------------------
!
      FUNCTION Tx(a,x,N2)
      INTEGER, INTENT(IN) :: N2 
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN) :: a, x
      COMPLEX(KIND=DP), DIMENSION(SIZE(x)) :: Tx

      INTEGER :: n
      REAL(KIND=DP), DIMENSION(4*N2+15) :: wsave
      COMPLEX(KIND=DP), DIMENSION(N2) :: fftx, fftxflip, fftC, fftR
      COMPLEX(KIND=DP), DIMENSION(N2) :: Cxeval, Rxeval

      EXTERNAL dcffti, dcfftf, dcfftb

      n = Size(x)
      CALL dcffti(N2,wsave)
      fftx = 0.0_DP
      fftxflip = 0.0_DP
      fftC = 0.0_DP
      fftR = 0.0_DP
      fftx(1:n) = x
      fftxflip(1:n) = x(n:1:-1)
      fftC(1:n) = a(n:2*n-1)
      fftR(1:n) = a(n:1:-1)
      fftR(1) = 0.0_DP
      CALL dcfftf(N2,fftx,wsave)
      CALL dcfftf(N2,fftxflip,wsave)
      CALL dcfftf(N2,fftC,wsave)
      CALL dcfftf(N2,fftR,wsave)
      Cxeval = fftC * fftx
      Rxeval = fftR * fftxflip
      CALL dcfftb(N2,Cxeval,wsave)
      CALL dcfftb(N2,Rxeval,wsave)
      Tx = (Cxeval(1:n)+Rxeval(n:1:-1))/N2

      END FUNCTION Tx
!
!--------------------------------------------------------------------------
!

      FUNCTION InvToeplitz(u,v,b,n,capN)
!
! b DIMENSION(n)
      INTEGER, INTENT(IN) :: n, capN
      COMPLEX(KIND=DP), DIMENSION(:), INTENT(IN) :: u, v, b
      COMPLEX(KIND=DP), DIMENSION(n) :: InvToeplitz

      INTEGER :: N2, i, N2n
      INTEGER, DIMENSION(capN) :: indexD
      REAL(KIND=DP), DIMENSION(4*capN+15) :: wsave
      REAL(KIND=DP)                       :: fx
      COMPLEX(KIND=DP), DIMENSION(capN)   :: x, h1, h2, D

      EXTERNAL dcffti, dcfftf, dcfftb

      N2 = 2*capN

      CALL dcffti(capN,wsave)
      x = 0.0_DP
      x(1:n) = Conjg(b)
      CALL dcfftf(capN,x,wsave)
      x = Conjg(x)
      N2n = (N2-n) * 2
      indexD(1) = 0
      DO i = 1, capN-1
        indexD(i+1) = MOD(indexD(i) + N2n, N2)
      END DO
      indexD = indexD + 1

      D = Gz(indexD)
      x = D * x
      h1 = v(1:N2-1:2) * x
      h2 = u(1:N2-1:2) * x
      CALL dcfftf(capN,h1,wsave)
      CALL dcfftf(capN,h2,wsave)
      h1 = Conjg(Gz(2:capN+1) * h1)
      h2 = Conjg(Gz(2:capN+1) * h2)
      CALL dcfftf(capN,h1,wsave)
      CALL dcfftf(capN,h2,wsave)
      h1 = Conjg(h1)
      h2 = Conjg(h2)
      h1 = u(2:N2:2) * h1
      h2 = v(2:N2:2) * h2
      x = h1 - h2
      CALL dcfftf(capN,x,wsave)
      fx = REAL(N2,DP)*REAL(capN,DP)
      x = x/CMPLX(fx,0_DP,DP)
      x = -Conjg(Gz(2:capN+1)) * x
      InvToeplitz = x(1:n)

      END FUNCTION InvToeplitz

END MODULE Toeplitz_Module
